from precise.skaters.portfoliostatic.diagportfactory import diagonal_portfolio_factory
from precise.skaters.portfoliostatic.weakportfactory import weak_portfolio_factory
import numpy as np
from functools import partial
from precise.skaters.portfolioutil.seriation import corr_seriation_portfolio_factory
from precise.skaters.covarianceutil.covfunctions import top_schur_complement, multiply_by_inverse, \
    bottom_schur_complement, to_symmetric, inverse_multiply
from precise.skaters.portfolioutil.portfunctions import portfolio_variance
from pprint import pprint

# Acca Dacca Portfolio method
#
# It's gonna rock you
#
# But a work in progress


def prc_acdc_weak_5(cov=None, pre=None):
    return schur_complement_portfolio_factory(port=weak_portfolio_factory, cov=cov, pre=pre, n_split=5)


def prc_acdc_weak_25(cov=None, pre=None):
    return schur_complement_portfolio_factory(port=weak_portfolio_factory, cov=cov, pre=pre, n_split=25)


def prc_acdc_diag_5(cov=None, pre=None):
    return schur_complement_portfolio_factory(port=diagonal_portfolio_factory, cov=cov, pre=pre, n_split=5)


def prc_acdc_diag_25(cov=None, pre=None):
    return schur_complement_portfolio_factory(port=diagonal_portfolio_factory, cov=cov, pre=pre, n_split=25)


def prc_acdc_unit_5(cov=None, pre=None):
    return schur_complement_portfolio_factory(port=unit_portfolio_factory, cov=cov, pre=pre, n_split=5)


def prc_acdc_unit_25(cov=None, pre=None):
    return schur_complement_portfolio_factory(port=unit_portfolio_factory, cov=cov, pre=pre, n_split=25)



HRP_PORT = [ prc_acdc_weak_5, prc_acdc_weak_25, prc_acdc_weak_5, prc_acdc_weak_25 ]


def even_split(cov, n_split=5)->(float,float):
    """
        This simple method is the default, for now, but I'm working on a better one
    """
    try:
        n_dim = np.shape(cov)[0]
    except IndexError:
        print('huh')
        raise ValueError('not good')
    if n_dim<=n_split:
        return (n_dim,0)
    else:
        n1 = int(n_dim / 2)
        n2 = n_dim - n1
        return n1, n2


def schur_complement_portfolio_factory(seriator=None, port=None, splitter=None, cov=None, pre=None, n_split=5, c:[float]=None):
    """
        This is not a standard technique, as far as I know.
        It is my attempt to unify two quite different approaches (top down versus optimization)
        As compared with hierarchical risk parity (say) this uses conditional covariances (schur complement)
        It is motivated by the unitary case:

    :param port:       The portfolio generator eventually used on small portfolios    port( cov ) -> [ float ]
    :param splitter:   A function taking cov matrix and returning (n1,n2), telling us how to bisect the problem
    :param c:          The coefficients in the constraint   w^T c = 1
                       (This "c" vector defaults to a vector of ones)
    :return:
    """

    if port is None:
        port = diagonal_portfolio_factory

    if splitter is None:
        splitter = partial( even_split, n_split=n_split )


    port_kwargs = {'splitter':splitter,'port':port,'c':c}
    return corr_seriation_portfolio_factory(seriator=seriator, port=_acca_dacca_portfolio, port_kwargs=port_kwargs, cov=cov, pre=pre)


def _acca_dacca_portfolio(cov, port, splitter):
    """
        Assumes assets have been ordered to try to reduce the importance of "B"
        (How best to do that remains an open question)

    """
    n1, n2 = splitter(cov=cov)
    assert n1+n2 == np.shape(cov)[0]
    if n1==0 or n2==0:
        return port(cov)
    else:
        #           cov   = [ [ A, B ]
        #                   [ C, D ] ]
        n = n1+n2
        A = cov[:n1,:n1]
        B = cov[:n1,n1:]
        C = cov[n1:,:n1]
        D = cov[n1:,n1:]
        Ac = to_symmetric( top_schur_complement(A=A, B=B, C=C, D=D ) )
        Dc = to_symmetric( bottom_schur_complement(A=A, B=B, C=C, D=D) )

        B_Dinv = multiply_by_inverse(B,D)  # B D^{-1}
        C_Ainv = multiply_by_inverse(C,A)  # B A^{-1}
        R = np.block([
                        [ np.eye(n1),        -B_Dinv ],
                        [ -C_Ainv    ,        np.eye(n2) ]
                    ])


        VERIFY_MATRIX_IDENTITY = True
        if VERIFY_MATRIX_IDENTITY:
            # debuggin'
            C_tilde = np.block( [
                          [ Ac,  np.zeros(shape=(n1,n2)) ],
                          [ np.zeros(shape=(n2,n1)), Dc  ]
            ])
            C_inv = inverse_multiply(C_tilde,R)
            pre = np.linalg.inv(cov)
            assert np.allclose( C_inv, pre )

        c = np.ones(n)
        c_tilde = np.dot(R, c)
        cA_tilde = c_tilde[:n1]
        cD_tilde = c_tilde[n1:]

        A_tilde = Ac
        D_tilde = Dc

        # Solve transformed problem
        wA_tilde = _acca_dacca_portfolio(cov=Ac, port=port, splitter=splitter)
        wD_tilde = _acca_dacca_portfolio(cov=Dc, port=port, splitter=splitter)

        # Allocate based on unitary portfolios
        from precise.skaters.portfoliostatic.unitportfactory import rescaled_unit_from_cov
        uA = rescaled_unit_from_cov(cov=Ac, c=cA_tilde)
        uD = rescaled_unit_from_cov(cov=Dc, c=cD_tilde)

        # this is all wrong, need to fix it

        massA = np.sum( np.linalg.solve(Ac,cA_tilde) )
        massB = np.sum( np.linalg.solve(Dc,cD_tilde) )

        varA = portfolio_variance(cov=Ac, w=wA_tilde)
        varD = portfolio_variance(cov=Dc, w=wD_tilde)
        cA_tilde = cA_tilde/varA
        cD_tilde = cD_tilde /varD
        c_tilde = np.concatenate([cA_tilde, cD_tilde])
        w_raw = np.linalg.solve(R, c_tilde)       # <-- But now diagonal terms come into play again
        assert np.allclose( np.dot(R,w_raw), c_tilde )
        from precise.skaters.locationutil.vectorfunctions import normalize
        w = normalize(w_raw)

        CURIOUS = False
        if CURIOUS:
            # Just for interest, check the impact of diagonal terms
            massA = np.sum(w[:n1])/np.sum(w)
            diagMassA = (1/varA)/(1/varA+1/varD)
            from precise.skaters.portfoliostatic.diagportfactory import prc_diag_alloc
            invMassA, invMassD = prc_diag_alloc([A,D])
            invMassA_tilde, invMassB_tilde = prc_diag_alloc([Ac, Dc])

            aAllocations = {'inv':invMassA,'inv_cond':invMassA_tilde,
                            'diag':diagMassA,'schur':massA}
            from pprint import pprint
            pprint(aAllocations)
            pass

        return w


if __name__=='__main__':
    from precise.skatertools.syntheticdata.factor import create_disjoint_factor_dataset, create_factor_dataset
    x = create_disjoint_factor_dataset(n_dims=[5,5], n=100)
    x = create_factor_dataset(n=200,n_dim=8)
    cov = np.cov(x, rowvar=False)
    from precise.skaters.covarianceutil.covfunctions import cov_to_corrcoef
    corr = cov_to_corrcoef(cov)

    # Compare in-sample portfolio variance
    from precise.skaters.portfoliostatic.schurportfactory import prc_hrp_diag_n5_g0, prc_hrp_weak_n5_g0, prc_hrp_unit_n5_g0
    from precise.skaters.portfoliostatic.unitportfactory import unit_portfolio_factory

    uv = portfolio_variance(cov=cov, w=unit_portfolio_factory(cov=cov))
    report = {'unitary': portfolio_variance(cov=cov, w=unit_portfolio_factory(cov=cov)) / uv,
              'divide and condition (unit)':portfolio_variance(cov=cov,w=prc_acdc_unit_5(cov=cov))/uv,
              'divide and condition (diag)': portfolio_variance(cov=cov, w=prc_acdc_diag_5(cov=cov))/uv,
              'divide and condition (weak)': portfolio_variance(cov=cov, w=prc_acdc_weak_5(cov=cov))/uv,
              'hierarchical risk parity (diag)': portfolio_variance(cov=cov, w=prc_hrp_diag_n5_g0(cov=cov)) / uv,
              'hierarchical risk parity (weak)': portfolio_variance(cov=cov, w=prc_hrp_weak_n5_g0(cov=cov)) / uv,
              'hierarchical risk parity (unit)': portfolio_variance(cov=cov, w=prc_hrp_unit_n5_g0(cov=cov)) / uv,
              }
    print(' ')
    pprint(report)



from precise.skatertools.syntheticdata.factor import create_band_dataset, create_factor_dataset
import numpy as np

# Generate random covariance matrices

def random_factor_cov(n=500, n_dim=100):
    xs = create_factor_dataset(n=n, n_dim=n_dim)
    return np.cov(xs,rowvar=False)


def random_band_cov(n=250, n_dim=20, n_bands=5):
    xs = create_band_dataset(n=250, n_dim=20, n_bands=5)
    return np.cov(xs, rowvar=False)



import numpy as np


def d1_factory( f, y, s, k=1, **kwargs ):
    """
         Skaters where changes to y are assumed iid

    :param f:   skater producing estimates of changes
    :param kwargs:   additional kwargs to f
    :return:
    """
    if not s or s.get('dy'):
        s = {'prev_y': y,
             'dy': {}}
        return y, np.eye(len(y)), s
    else:
        dy = y - s['prev_y']
        dy_hat, dy_cov, s['dy'] = f(y=dy, s=s['dy'], k=k, **kwargs)
        y['prev_y'] = y
        x = s['prev_y'] + dy_hat
        return x, dy_cov, s


from precise.skaters.managers.ppomanagerfactory import ppo_ewa_long_manager_factory, ppo_pm_long_manager_factory, ppo_long_manager_factory
from precise.skaters.covariance.bufsk import buf_sk_glcv_pcov_d0_n100, buf_sk_glcv_pcov_d0_n100_t0, buf_sk_ld_pcov_d0_n100, buf_sk_mcd_pcov_d0_n100, buf_sk_ld_pcov_d1_n100

# PyPortfolioOpt managers

def ppo_pm_t0_d0_r025_n50_vol_long_manager(y, s, k=1, e=1):
    assert k == 1
    return ppo_pm_long_manager_factory(y=y, s=s, method='min_volatility', target=0, e=e, r=0.025, n_emp=50)


def ppo_ewa_d0_r025_n50_vol_long_manager(y, s, k=1, e=1):
    assert k == 1
    return ppo_ewa_long_manager_factory(y=y, s=s, method='min_volatility', e=e, r=0.025, n_emp=50)


def ppo_pm_t0_d0_r025_n50_quad_long_manager(y, s, k=1, e=1):
    assert k == 1
    return ppo_pm_long_manager_factory(y=y, s=s, method='max_quadratic_utility', target=0, e=e, r=0.025, n_emp=50)


def ppo_ewa_d0_r025_n50_quad_long_manager(y, s, k=1, e=1):
    assert k == 1
    return ppo_ewa_long_manager_factory(y=y, s=s, method='max_quadratic_utility', e=e, r=0.025, n_emp=50)


def ppo_pm_t0_d0_r025_n50_sharpe_long_manager(y, s, k=1, e=1):
    assert k == 1
    return ppo_pm_long_manager_factory(y=y, s=s, method='max_sharpe', target=0, e=e, r=0.025, n_emp=50)


def ppo_ewa_d0_r025_n50_sharpe_long_manager(y, s, k=1, e=1):
    assert k == 1
    return ppo_ewa_long_manager_factory(y=y, s=s, method='max_sharpe', e=e, r=0.025, n_emp=50)


# Sklearn with min vol

def ppo_sk_ld_pcov_d1_n100_vol_long_manager(y, s, k=1, e=1):
    return ppo_long_manager_factory(y=y,s=s,f=buf_sk_ld_pcov_d0_n100, e=e, method='min_volatility')


def ppo_sk_glcv_pcov_d0_n100_vol_long_manager(y, s, k=1, e=1):
    return ppo_long_manager_factory(y=y,s=s,f=buf_sk_glcv_pcov_d0_n100, e=e, method='min_volatility')


def ppo_sk_glcv_pcov_d0_n100_t0_vol_long_manager(y, s, k=1, e=1):
    return ppo_long_manager_factory(y=y,s=s,f=buf_sk_glcv_pcov_d0_n100_t0, e=e, method='min_volatility')


def ppo_sk_mcd_pcov_d0_n100_vol_long_manager(y, s, k=1, e=1):
    return ppo_long_manager_factory(y=y,s=s,f=buf_sk_mcd_pcov_d0_n100, e=e, method='min_volatility')


# Sklearn with quadratic util

def ppo_sk_ld_pcov_d1_n100_quad_long_manager(y, s, k=1, e=1):
    return ppo_long_manager_factory(y=y,s=s,f=buf_sk_ld_pcov_d0_n100, e=e, method='max_quadratic_utility')


def ppo_sk_glcv_pcov_d0_n100_quad_long_manager(y, s, k=1, e=1):
    return ppo_long_manager_factory(y=y,s=s,f=buf_sk_glcv_pcov_d0_n100, e=e, method='max_quadratic_utility')


def ppo_sk_glcv_pcov_d0_n100_t0_quad_long_manager(y, s, k=1, e=1):
    return ppo_long_manager_factory(y=y,s=s,f=buf_sk_glcv_pcov_d0_n100_t0, e=e, method='max_quadratic_utility')


def ppo_sk_mcd_pcov_d0_n100_quad_long_manager(y, s, k=1, e=1):
    return ppo_long_manager_factory(y=y,s=s,f=buf_sk_mcd_pcov_d0_n100, e=e, method='max_quadratic_utility')



# Sklearn with max sharpe

def ppo_sk_ld_pcov_d1_n100_sharpe_long_manager(y, s, k=1, e=1):
    return ppo_long_manager_factory(y=y,s=s,f=buf_sk_ld_pcov_d0_n100, method='max_sharpe')


def ppo_sk_glcv_pcov_d0_n100_sharpe_long_manager(y, s, k=1, e=1):
    return ppo_long_manager_factory(y=y,s=s,f=buf_sk_glcv_pcov_d0_n100, method='max_sharpe')


def ppo_sk_glcv_pcov_d0_n100_t0_sharpe_long_manager(y, s, k=1, e=1):
    return ppo_long_manager_factory(y=y,s=s,f=buf_sk_glcv_pcov_d0_n100_t0, method='max_sharpe')


def ppo_sk_mcd_pcov_d0_n100_sharpe_long_manager(y, s, k=1, e=1):
    return ppo_long_manager_factory(y=y,s=s,f=buf_sk_mcd_pcov_d0_n100, method='max_sharpe')



PPO_LONG_MANGERS = [ppo_pm_t0_d0_r025_n50_vol_long_manager,
                    ppo_ewa_d0_r025_n50_vol_long_manager,
                    ppo_pm_t0_d0_r025_n50_quad_long_manager,
                    ppo_ewa_d0_r025_n50_quad_long_manager,
                    ppo_ewa_d0_r025_n50_quad_long_manager,
                    ppo_pm_t0_d0_r025_n50_sharpe_long_manager,
                    ppo_ewa_d0_r025_n50_sharpe_long_manager,
                    ppo_sk_ld_pcov_d1_n100_vol_long_manager,
                    ppo_sk_glcv_pcov_d0_n100_vol_long_manager,
                    ppo_sk_glcv_pcov_d0_n100_t0_vol_long_manager,
                    ppo_sk_mcd_pcov_d0_n100_vol_long_manager,
                    ppo_sk_ld_pcov_d1_n100_quad_long_manager,
                    ppo_sk_glcv_pcov_d0_n100_quad_long_manager,
                    ppo_sk_glcv_pcov_d0_n100_t0_quad_long_manager,
                    ppo_sk_mcd_pcov_d0_n100_quad_long_manager,
                    ppo_sk_ld_pcov_d1_n100_sharpe_long_manager,
                    ppo_sk_glcv_pcov_d0_n100_sharpe_long_manager,
                    ppo_sk_glcv_pcov_d0_n100_t0_sharpe_long_manager,
                    ppo_sk_mcd_pcov_d0_n100_sharpe_long_manager]



from precise.skaters.portfoliostatic.unitallocfactory import unitary_min_var_allocation_factory


def unit_alloc(covs=None, pres=None):
    return unitary_min_var_allocation_factory(covs=covs,pres=pres)


UNIT_ALLOC = [unit_alloc]


import numpy as np
import pandas as pd
from precise.skatertools.m6.covarianceforecasting import m6_cov
from precise.skatertools.m6.tilting import affection_tilt

# M6 Quintile probability estimates by Monte Carlo


def what_pctl_number_of(x, a, pctls=[20,40,60,80]):
    return np.argmax(np.sign(np.append(np.percentile(x, pctls), np.inf) - a))


def mvn_quintile_probabilities(sgma, n_samples, mu=None):
    n_dim = np.shape(sgma)[0]
    if mu is None:
        mu = np.zeros(n_dim)
    x =  np.random.multivariate_normal(mu, sgma, size=n_samples, check_valid='warn', tol=1e-8)
    y = scores_to_quintiles(x)
    p = list()
    for i in range(5):
        pi = np.mean(y==i,axis=0)
        p.append(pi)
    return p


def scores_to_quintiles(x):
    ys = list()
    for xi in x:
        q = np.quantile(x,[0.2,0.4,0.6,0.8])
        y = np.searchsorted(q, xi)
        ys.append(y)
    return np.array(ys)



def m6_probabilities(f, interval='d',n_dim=100, n_samples=5000, n_obs=200, verbose=False, love=None, hate=None, intensity=1.0):
    if verbose:
        print('   ... retrieving data and estimating covariance')
    covdf = m6_cov(f=f, interval=interval, n_dim=n_dim, n_obs=n_obs)
    tickers = list(covdf.columns)
    sgma = covdf.values
    mu = affection_tilt(covdf=covdf, love=love, hate=hate, intensity=intensity)

    if verbose:
        print('   ... performing Monte Carlo for rank probabilities')

    p = mvn_quintile_probabilities(mu=mu, sgma=sgma, n_samples=n_samples)
    df = pd.DataFrame(columns=tickers, data=p).transpose()
    df_cov = pd.DataFrame(columns=tickers, index=tickers, data=sgma)
    return df, df_cov


if __name__=='__main__':
    df = m6_probabilities(n_dim=5)







import numpy as np
from precise.skaters.covariance.runempfactory import _emp_pcov_init, _emp_pcov_update
import math
from typing import Union, List

# Exponential weighted sample covariance


def ewa_emp_pcov_factory(y, s:dict, k=1, r=0.025, n_emp=None, e=1):
    assert k==1
    s = ema_scov(s=s,x=y,r=r, n_emp=n_emp)
    x = s['mean']
    x_cov = s['pcov']
    return x, x_cov, s


def ema_scov(s:dict, x:Union[List[float], int]=None, r:float=0.025, n_emp=None):
    """ Maintain running population covariance """
    if s.get('n_samples') is None:
        if isinstance(x,int):
            return _ema_scov_init(n_dim=x,r=r, n_emp=n_emp)
        elif isinstance(x,(List,np.ndarray)):
            s = _ema_scov_init(n_dim=len(x),r=r, n_emp=n_emp)
        else:
            raise ValueError('Not sure how to initialize EWA COV tracker. Supply x=5 say, for 5 dim')
    if x is not None:
        s = _ema_scov_update(s=s, x=x, r=r)
    return s


def _ema_scov_init(n_dim=None, r:float=0.025, n_emp=None ):
    """ Initialize object to track exp moving avg cov for zero mean

       r:       Importance of current data point
       n_emp:   Discouraged. Really only used for tests.
                This is the number of samples for which empirical is used, rather
                than running updates. By default n_emp ~ 1/r

    """
    if n_emp is None:
        n_emp = int(min(250, max(5, math.ceil(1 / r))))
    s = _emp_pcov_init(n_dim=n_dim)
    s.update({'rho':r, 'n_emp':n_emp})
    return s


def _ema_scov_update(s:dict, x:[float], r:float=None, target=None, y=None):
    """ Update recency weighted estimate of scov for zero mean

          If target is not None, it will be used in place of the mean when updating
          Obviously, this changes the interpretation of 'scov'

          xt - transpose of x

    """
    if s['n_samples']< s['n_emp']:
        # Use the regular cov update for a burn-in period
        # During this time both scov and pcov are maintained
        s = _emp_pcov_update(s=s, x=x, target=target)
        if s['n_samples']>1:
            s['scov'] = s['pcov'] * s['n_samples'] / (s['n_samples'] - 1)
    else:
        s['n_samples']+=1
        r = s['rho'] if r is None else r
        assert s['n_dim'] == len(x)
        xcol = np.ndarray(shape=(s['n_dim'], 1))
        if target is not None:
             xcol[:,0] = x - target
        else:
            xcol[:,0] = x - s['mean']
        if y is None:
            ycolT = xcol.T
        else:
            ycol = np.ndarray(shape=(s['n_dim'], 1))
            if target is not None:
                ycol[:, 0] = y - target
            else:
                ycol[:, 0] = y - s['mean']
            ycolT = ycol.T

        yyt = np.dot(xcol, ycolT)
        s['scov'] = (1 - r) * s['scov'] + r * yyt
        s['mean'] = (1 - r) * s['mean'] + r * x
        s['pcov']= s['scov']*(s['n_samples']-1)/s['n_samples']
    return s









from pprint import pprint
from precise.skatervaluation.battleutil.battleio import win_data
from precise.skatervaluation.battleutil.eloformulas import elo_change
from collections import Counter
import random
from precise.skatervaluation.battleutil.speed import TIMING
from collections import OrderedDict
# Creating Elo ratings from collections of wins and losses stored in hashed files /battleresults



def elo_from_win_files(genre='cov_likelihood'):
    """
    :return:  Elo ratings for all categories
    """
    # MAYBETODO: It would be easy to make this // across categories but not a high priority :)
    the_lot = [(cat, elo_from_win_counts(cat_data, timing_genre=genre)) for cat, cat_data in win_data(genre=genre)]
    return the_lot


def elo_from_win_counts(ctn, timing_genre=None):
    """
        Elo ratings from a counter or dict of match results

        Each key is a string of the form:

              someone>someoneelse

        indicating a win of someone over someonelese. Values are the number of occasions on
        which the result eventuated.

        Matches are sampled in random ordering, so the Elo ratings will be different each
        time this script is run.

        There is no temporality. When older battle results files become irrelevant they
        should simply be deleted
    """
    contestants = list(set( [ k.split('>')[0] for k in ctn.keys() ] + [ k.split('>')[1] for k in ctn.keys() ]))
    elo = Counter( dict([ (c,1500) for c in contestants ]))
    finished = False
    while not finished:
        remaining_counts = [ c for b,c in ctn.items() if c>=1 ]
        remaining_battles = [ (b,c) for b,c in ctn.items() if c>=1 ]
        weights = [ c/sum(remaining_counts) for c in remaining_counts ]
        finished = not remaining_battles
        if not finished:
            random_battle = random.choices(population=remaining_battles, weights=weights,k=1)[0]
            winner, loser = random_battle[0].split('>')
            winner_change, loser_change = elo_change(elo[winner],elo[loser],points=1.0, k=10)
            ctn[random_battle[0]] -= 1
            elo[winner] += winner_change
            elo[loser] += loser_change
    if timing_genre is not None:
        contestant_timing = TIMING.get(timing_genre)
        if contestant_timing is not None:
            with_timing = sorted( [ (contestant,(score,contestant_timing.get(contestant))) for contestant, score in elo.items()], key= lambda x: x[1][0], reverse=True)
            elo = OrderedDict(with_timing)
    return elo



if __name__=='__main__':
   pprint(elo_from_win_files(genre='manager_var'))


from precise.skaters.portfolioutil.portfunctions import portfolio_variance
from precise.skaters.portfoliostatic.unitportfactory import unit_portfolio_factory
from typing import List
from itertools import zip_longest
from precise.skaters.locationutil.vectorfunctions import normalize



def unitary_portfolio_variance(cov=None, pre=None, **kwargs):
    """
        Variance of the unit min-var portfolio
        (Used in some hierarchical methods to allocate capital)
    """
    w = unit_portfolio_factory(pre=pre, cov=cov, **kwargs)
    return portfolio_variance(cov=cov,w=w)


def unitary_min_var_allocation_factory(covs:List, pres:List=None)->[float]:
    """ Allocate capital between portfolios using either cov or pre matrices
    :param covs:  List of covariance matrices
    :param pres:  List of precision matrices
    :return: Capital allocation vector
    """
    if pres is None:
        pres = []
    return normalize([ 1/unitary_portfolio_variance(cov=cov, pre=pre) for cov, pre in zip_longest(covs, pres, fillvalue=None) ])








import numpy as np
from precise.skaters.covarianceutil.conventions import X_TYPE, X_DATA_TYPE, is_data
from precise.skaters.covarianceutil.datafunctions import data_population_covariance, data_population_correlation

# State machines that track stats for finite buffers of vectors


def buf_pcov_factory(func, y:X_TYPE=None, s:dict=None, n_buffer:int=100, e=1):
    """
         Factory for building estimators from functions that return dicts with 'loc' and 'pcov' keys

    :param func:  Acts on buffer and returns object with 'loc', 'pcov'
    :param y:
    :param s:
    :param n_buffer:
    :return:
    """
    s = _buf(funcs=[func], func_names=['mav'], func_kwargs=[{}],s=s, x=y, n_buffer=n_buffer, e=e)
    if s.get('mav') is not None:
        x = np.array(s['mav']['loc'])
        x_cov = np.array(s['mav']['pcov'])
    else:
        x = y
        x_cov = np.eye(len(y))
    return x, x_cov, s


def buf_cov(s:dict=None, x:X_TYPE=None, n_buffer:int=100)->dict:
    # Equivalent to np.nanstd(xs[max(0, k - n_buffer + 1):k + 1], axis=0)
    return _buf1(func=data_population_correlation, func_name='corrcoef', s=s, x=x, n_buffer=n_buffer)


def buf_std(s:dict=None, x:X_TYPE=None, n_buffer:int=100)->dict:
    # Equivalent to np.nanstd(xs[max(0, k - n_buffer + 1):k + 1], axis=0)
    return _buf1(func=np.nanstd, func_name='mean', s=s, x=x, n_buffer=n_buffer, axis=0)


def buf_mean(s:dict=None, x:X_TYPE=None, n_buffer:int=100)->dict:
    # Equivalent to np.nanmean(xs[max(0, k - n_buffer + 1):k + 1], axis=0)
    return _buf1(func=np.nanmean, func_name='mean', s=s, x=x, n_buffer=n_buffer, axis=0)


def buf_median(s:dict=None, x:X_TYPE=None, n_buffer:int=100)->dict:
    # Equivalent to np.nanmedian(xs[max(0, k - n_buffer + 1):k + 1], axis=0)
    return _buf1(func=np.nanmedian, func_name='median', s=s, x=x, n_buffer=n_buffer, axis=0)


def _buf1(func, func_name, s:dict=None, x:X_TYPE=None, n_buffer:int=100, e=1, **func_kwarg)->dict:
    return _buf(funcs=[func], func_names=[func_name], func_kwargs=[func_kwarg], s=s, x=x, n_buffer=n_buffer,e=e)


def buf_mean_and_median(s:dict=None, x:X_TYPE=None, n_buffer:int=100, e=1)->dict:
    return _buf(funcs=[np.nanmean, np.nanmedian], func_names=['mean','median'], func_kwargs=[{'axis':0},{'axis':0}], s=s, x=x, n_buffer=n_buffer, e=e)


def buf_mean_and_pcov(s:dict=None, x:X_TYPE=None, n_buffer:int=100, e=1)->dict:
    # Equivalent to np.cov(xs[max(0, k - n_buffer + 1):k + 1], axis=0)
    return _buf(funcs=[np.nanmean, data_population_covariance], func_names=['mean', 'pcov'], func_kwargs=[{'axis':0}, {}], s=s, x=x, e=e, n_buffer=n_buffer)


def _buf(funcs, func_names:[str], func_kwargs:[dict], s:dict=None, x:X_TYPE=None, n_buffer:int=100, e=1)->dict:
    """
    :param funcs:            [func(xs, axis=0) -> 1d array]
    :param func_names:
    :param s:
    :param x:
    :param n_buffer:
    :param e:             If e>1 calculation is performed
    :return:
    """
    if not s:
        s = _buf_init(s=s, n_buffer=n_buffer)
    if is_data(x):
        s = _buf_update(funcs=funcs, func_names=func_names, func_kwargs=func_kwargs, s=s, x=x, n_buffer=n_buffer, e=e)
    return s


def _buf_init(s:dict=None, n_buffer:int=None)->dict:
    if not s:
        s = dict()
    s.update( {'buffer':list(), 'n_buffer':n_buffer} )
    return s


def _buf_update(funcs, func_names, func_kwargs, s:dict, x:X_DATA_TYPE, n_buffer:int=None, e=1)->dict:
    if n_buffer is None:
        n_buffer = s['n_buffer']
    if len(s['buffer']):
        assert len(s['buffer'][0])==len(x), 'dimension mismatch'
    s['buffer'].append(np.atleast_1d(x))
    if len(s['buffer'])>n_buffer:
        _ = s['buffer'].pop(0)
    for func, func_name, func_kwargs in zip(funcs, func_names, func_kwargs):
        if e>0:
            s[func_name] = func(s['buffer'],**func_kwargs)
    return s




if __name__=='__main__':
    import random
    xs = np.random.randn(500,100)
    xs[17,1] = np.nan
    s = {}
    n_buffer = random.choice([1,5,100])
    for k,x in enumerate(xs):
        s = buf_mean_and_median(s=s, x=x, n_buffer=n_buffer)
        np_median = np.nanmedian(xs[max(0,k-n_buffer+1):k+1], axis=0)
        s_median = s['median']
        if not np.allclose( np_median, s_median, atol=1e-8 ):
            print('urgh')
            assert False



import numpy as np

def symmetric_step_up_matrix(n1, n2):
    """
       :return M such that M np.ones(n2) = np.ones(n1)
    """

    def _Mj(n1, n2, j):
        assert abs(n1 - n2) <= 1
        nd = n2 - n1
        if nd == 0:
            return np.eye(n1)
        else:
            j_row = np.ones(shape=(1, n2)) / n2
            E = np.eye(n2)
            Mj = np.concatenate([E[:j], j_row, E[j:]], axis=0)
            assert np.linalg.norm(np.dot(Mj, np.ones(n2)) - np.ones(n1)) < 1e-8
            return Mj

    if n1 <n2:
        M = symmetric_step_up_matrix(n2, n1).transpose() * n1 / n2
    elif n1==n2:
        M = np.eye(n1)
    else:
        M = np.zeros(shape=(n1 ,n2))
        for j in range(n1):
            M = M + _Mj(n1 ,n2 ,j ) /n1
    assert np.linalg.norm(np.dot(M, np.ones(n2)) - np.ones(n1)) < 1e-8
    return M


def even_split(cov, n_split=5)->(float,float):
    """
        Parity methods require some way of splitting assets into two groups
        This simple method is the default
    """
    n_dim = np.shape(cov)[0]
    if n_dim<=n_split:
        return (n_dim,0)
    else:
        n1 = int(n_dim / 2)
        n2 = n_dim - n1
        return n1, n2


from precise.skaters.covariance.allcovskaters import ALL_D0_SKATERS
from precise.skaters.managers.allmanagers import LONG_MANAGERS
from precise.skaters.covarianceutil.likelihood import cov_likelihood
from precise.skaters.managerutil.managerstats import manager_info, manager_var
from precise.skatervaluation.battledata.allsources import params_category_and_data
from uuid import uuid4
import os
import json
import pathlib
from pprint import pprint
import traceback
from collections import Counter
from momentum.functions import rvar
from precise.whereami import BATTLE_RESULTS_DIR
import numpy as np
import time


def manager_info_battle(params:dict):
    return generic_battle(contestants=LONG_MANAGERS, evaluator=manager_info, params=params, atol=1e-8)


def manager_var_battle(params:dict):
    return generic_battle(contestants=LONG_MANAGERS, evaluator=manager_var, params=params, atol=1e-8)


def cov_likelihood_battle(params:dict):
    contestants = ALL_D0_SKATERS
    return generic_battle(contestants=contestants, evaluator=cov_likelihood, params=params, atol=1.0)


def generic_battle(contestants, evaluator, params:dict, atol=1.0):
    """
        Write results to a new queue.
        evaluator(contestant=contestant, xs=xs, n_burn=params['n_burn'], with_metrics=True, lb=lb, ub=ub)
    """
    evaluator_name = evaluator.__name__
    n_per_battle = 5
    try:
        params, category, xs_test = params_category_and_data(params=params)
    except Exception as e:
        print(e)
        pprint(params)
        print('Something is probably wrong with params for getting data, so this config will not fly')
        params, category, xs_test = params_category_and_data(params=params)

    print('Data retrieval test passed for category '+category)
    pprint(params)
    time.sleep(1)
    print('Will test the following contestants')
    pprint(contestants)

    qn = str(uuid4())+'.json'
    queue_dir = os.path.join(BATTLE_RESULTS_DIR, evaluator_name, category)
    queue = os.path.join(queue_dir,qn)
    pathlib.Path(queue_dir).mkdir(parents=True, exist_ok=True)
    print(queue)

    battles = Counter()
    timing = dict()
    reliability = dict()
    failures = dict()

    worst_assessment_seen = 10000000
    lb = params['lb']
    ub = params['ub']

    while True:
        n_obs = params['n_obs']
        params, category, xs = params_category_and_data(params=params)
        assert len(xs)==n_obs
        xs = np.array(xs)
        np.random.shuffle(contestants)
        some_contestants = contestants[:n_per_battle]

        stuff = list()
        for contestant in some_contestants:
            try:
                if 'manager' in contestant.__name__ or True:
                    print('  '+contestant.__name__)
                assessment, metrics = evaluator(contestant=contestant, xs=xs, n_burn=params['n_burn'], lb=lb, ub=ub)
                metrics['name']=contestant.__name__
                metrics['traceback']=''
                metrics['passing']=1
                if assessment<worst_assessment_seen:
                    worst_assessment_seen = assessment
                    print({'worst_assessment_yet':assessment})
                name = metrics['name']
                if name not in timing:
                    timing[name] = {}
                timing[name] = rvar(timing[name], x=metrics['time'], rho=0.05)
                if name not in reliability:
                    reliability[name] = {}
                reliability[name] = rvar(reliability[name], x=1.0, rho=0.05)
            except Exception as e:
                metrics = {'name':contestant.__name__,'passing':0,'traceback':traceback.format_exc(),'ll':-100000000}
                if contestant.__name__ not in reliability:
                    reliability[contestant.__name__] = {}
                reliability[contestant.__name__] = rvar(reliability[contestant.__name__], x=0.0, rho=0.05)
                failures[contestant.__name__] = traceback.format_exc()
                assessment = worst_assessment_seen
            stuff.append( (assessment,metrics))
        valid = [ s for s in stuff if s[1]['passing']>0.5 ]

        if len(valid)<=2:
            print('Less than 2 working contestants this time around: ')
            pprint(some_contestants)
            for contestant in some_contestants:
                pprint(failures.get(contestant.__name__))
            print('Urgh')

        for i, mi in enumerate(valid):
            for j, mj in enumerate(valid):
                if j != i:
                    if mi[0] > mj[0]+atol:
                        i_name = mi[1]['name']
                        j_name = mj[1]['name']
                        cmp_name = i_name+'>'+j_name
                        battles.update({cmp_name:1.0})

        reliabilties = dict([(nm, reliab['mean']) for nm,reliab in reliability.items() ] )
        cpu_times = dict([(nm, tm['mean']) for nm, tm in timing.items()])

        if np.random.rand()<1:
            with open(queue,'wt') as fh:
                print('Saving')
                json.dump(battles,fh)
                print('---')
                pprint(reliabilties)
                print('---')
                pprint(battles)
                print(' ')
                pprint(failures)
                print('---')
                pprint(cpu_times)
















from precise.skaters.portfoliostatic.equalport import equal_long_port


def static_cov_manager_factory_d0(y, s, f, port, e=1, f_kwargs:dict=None, port_kwargs:dict=None, n_cold=5):
    """
       Basic manager pattern ignoring mean.
       Expects to receive changes in log(price).
       If you have opinions on means, you'll have to incorporate them into variance somehow
       via the covariance skater. 

          :param f     cov skater ("d0")
          :param port  portfolio constructor

       :returns w, s
    """

    if f_kwargs is None:
        f_kwargs = {}
    if port_kwargs is None:
        port_kwargs = {}
    if not s:
        s = {'f_state':{},
             'port_state':{},
             'count':0}

    x_mean, x_cov, s['f_state'] = f(y=y,s=s['f_state'], k=1, e=e, **f_kwargs)
    s['count']+=1
    if s['count']>=n_cold and (e>0):
        w = port(cov=x_cov, **port_kwargs)
    else:
        w = equal_long_port(cov=x_cov)
    return w, s




import numpy as np
from precise.skaters.covariance.ewaempfactory import _ema_scov_update, _ema_scov_init
from typing import Union, List
from precise.skaters.location.averagingpre import sma

# Exponential weighted partial moment covariance estimation

# If no target is supplied, either initially or for the update call, then a running mean will be used.

QUADRANTS = {'cu':(1.0,1,1),    # x*1 > 0  y*1 > 0
             'du':(-1.0,-1,1),
             'dl':(-1.0,1,-1),
             'cl':(1.0,-1,-1)}


def ewa_pm_factory(s, y, k, r, target=0, n_emp=None, e=1):
    """ Skater """
    assert k==1
    s = partial_ema_scov(s=s,x=y,r=r,target=target)
    x = s['sma']['mean']
    x_cov = s['scov']
    return x, x_cov, s


def partial_ema_scov(s:dict, x:Union[List[float], int]=None, r:float=0.025, target=None, n_emp=None):
    """ Maintain running population covariance """
    if s.get('n_samples') is None:
        if isinstance(x,(int,float)):
            return _partial_ema_scov_init(n_dim=int(x), r=r, target=target, n_emp=n_emp)
        elif len(x)>1:
            s = _partial_ema_scov_init(n_dim=len(x), r=r, target=target, n_emp=n_emp)
        else:
            raise ValueError('Not sure how to initialize EWA COV tracker. Supply x=5 say, for 5 dim')
    if x is not None:
        s = _partial_ema_scov_update(s=s, x=x, r=r)
    return s


def _partial_ema_scov_init(n_dim=None, r:float=0.025, n_emp=None, target:float=None)->dict:
    """ Initialize object to track partial moments

       r:       Importance of current data point
       n_emp:   Discouraged. Really only used for tests.
                This is the number of samples for which empirical is used, rather
                than running updates. By default n_emp ~ 1/r

    """
    s = dict([ (q,_ema_scov_init(n_dim=n_dim,r=r,n_emp=n_emp)) for q in QUADRANTS ])
    q = next(iter(s.keys())) # Choose any
    s['n_dim'] = s[q]['n_dim']
    s['n_emp'] = s[q]['n_emp']
    s['rho'] = s[q]['rho']
    s['target'] = target
    s['sma'] = sma({},n_dim,r=r)
    return s


def _partial_ema_scov_update(s:dict, x:[float], r:float=None, target=None):
    """ Update recency weighted estimate of scov-like matrix by treating quadrants individually """

    assert len(x)==s['n_dim']

    # If target is not supplied we maintain a mean that switches from emp to ema
    if target is None:
        target = s['target']
    if target is None:
        target = s['sma']['mean']

    # Update running partial scatter estimates
    for q,(w,sgn1,sgn2) in QUADRANTS.items():
        # Morally:
        #    x1 = max(0, (x-target)*sgn1) * sgn1
        #    x2 = (np.max(0, (x-target)*sgn2) * sgn2) if sgn1!=sgn2 else x1
        x1 = (x-target)*sgn1
        x2 = (x-target)*sgn2
        x1[x1<0]=0
        x2[x2<0]=0
        x1 = sgn1*x1
        x2 = sgn2*x2
        s[q] = _ema_scov_update(s[q],x=x1,r=r,target=0, y=x2)

    s['mean'] = np.copy( s['sma']['mean'] )
    s['n_samples'] = s['sma']['n_samples']

    if s['n_samples']>=2:
        s['scov'] = np.zeros(shape=((s['n_dim'],s['n_dim'])))
        for q in QUADRANTS:
            try:
                s['scov'] += s[q]['scov']
            except:
                pass
    else:
        s['scov'] = np.eye(s['n_dim'])

    s['sma'] = sma(s=s['sma'], x=x, r=r)
    return s









import numpy as np





def parallel_bisection_root_finder(f, lb, ub, atol=1e-6, n_iter=10, copy=True, guess=None, *args, **kwargs):
    """  Find many zeros of f for many different choices of function parameters

        f:  function acting on 1d arrays return 1d arrays
        lb: (n_vars,) lower bounds
        ub: (n_vars,) upper bounds
        *args, **kwargs   Additional arguments for f that, we presume, vary from one variable to the next
        :returns vector of roots, vector of function vals (close to zero)
    """

    # Ack: https://stackoverflow.com/questions/13088115/finding-the-roots-of-a-large-number-of-functions-with-one-variable
    x0 = np.copy(lb) if copy else lb
    x1 = np.copy(ub) if copy else ub
    converged = False
    for i in range(n_iter):
        if i==0 and guess is not None:
            assert len(guess)==len(x0)
            x_mid = guess
        else:
            x_mid = (x0 + x1)/2.0
        f0 = f(x0, *args, **kwargs)
        f1 = f(x1, *args, **kwargs)
        f_mid = f(x_mid, *args, **kwargs)
        x0 = np.where( np.sign(f_mid) == np.sign(f0), x_mid, x0 )
        x1 = np.where( np.sign(f_mid) == np.sign(f1), x_mid, x1 )
        error_max = np.amax(np.abs(x1 - x0))
        if error_max < atol:
            converged = True
            break
    fraction_converged = np.mean(np.abs(x1 - x0) < atol)
    return (x0+x1)/0.5, fraction_converged



from precise.skaters.managers.schurmanagerfactory import schur_weak_pm_manager_factory, schur_weak_ewa_manager_factory


def schur_weak_pm_t0_d0_r025_n50_g100_long_manager(y, s, k=1,e=1):
    assert k==1
    return schur_weak_pm_manager_factory(y=y, s=s, target=0, e=e, r=0.025, n_emp=50, gamma=1.0)


def schur_weak_pm_t0_d0_r050_n25_g100_long_manager(y, s, k=1,e=1):
    assert k==1
    return schur_weak_pm_manager_factory(y=y, s=s, target=0, e=e,r=0.050, n_emp=25, gamma=1.0)


def schur_weak_emp_d0_r025_n50_g100_long_manager(y, s, k=1,e=1):
    assert k==1
    return schur_weak_ewa_manager_factory(y=y, s=s, e=e,r=0.025, n_emp=50, gamma=1.0)


def schur_weak_emp_d0_r050_n25_g100_long_manager(y, s, k=1,e=1):
    assert k==1
    return schur_weak_ewa_manager_factory(y=y, s=s, e=e,r=0.05, n_emp=25, gamma=1.0)


# gamma = 0.5

def schur_weak_pm_t0_d0_r025_n50_g050_long_manager(y, s, k=1,e=1):
    assert k==1
    return schur_weak_pm_manager_factory(y=y, s=s, target=0, e=e,r=0.025, n_emp=50, gamma=0.5)


def schur_weak_pm_t0_d0_r050_n25_g050_long_manager(y, s, k=1,e=1):
    assert k==1
    return schur_weak_pm_manager_factory(y=y, s=s, target=0, e=e,r=0.050, n_emp=25, gamma=0.5)


def schur_weak_emp_d0_r025_n50_g050_long_manager(y, s, k=1,e=1):
    assert k==1
    return schur_weak_ewa_manager_factory(y=y, s=s, e=e,r=0.025, n_emp=50, gamma=0.5)


def schur_weak_emp_d0_r050_n25_g050_long_manager(y, s, k=1,e=1):
    assert k==1
    return schur_weak_ewa_manager_factory(y=y, s=s, e=e,r=0.05, n_emp=25, gamma=0.5)



SCHUR_100_LONG_MANAGERS = [ schur_weak_pm_t0_d0_r025_n50_g100_long_manager,
                      schur_weak_pm_t0_d0_r050_n25_g100_long_manager,
                      schur_weak_emp_d0_r025_n50_g100_long_manager,
                      schur_weak_emp_d0_r050_n25_g100_long_manager
                      ]
SCHUR_50_LONG_MANAGERS = [ schur_weak_pm_t0_d0_r025_n50_g050_long_manager,
                      schur_weak_pm_t0_d0_r050_n25_g050_long_manager,
                      schur_weak_emp_d0_r025_n50_g050_long_manager,
                      schur_weak_emp_d0_r050_n25_g050_long_manager
                      ]

SCHUR_LONG_MANAGERS = SCHUR_100_LONG_MANAGERS + SCHUR_50_LONG_MANAGERS
SCHUR_LS_MANAGERS = []
SCHUR_MANAGERS = SCHUR_LONG_MANAGERS + SCHUR_LS_MANAGERS


import numpy as np 
from precise.skaters.locationutil.vectorfunctions import normalize
import pandas as pd


def equal_long_port(cov=None, pre=None, as_dense=False):
    """
      Equal weighted portfolio
      :param as_dense   If True, will always return np.array even if cov or pre
    """
    if pre is not None:
        n_dim = np.shape(pre)[0]
    else:
        n_dim = np.shape(cov)[0]
    w = np.array(normalize(np.ones(n_dim)))

    as_series = (not as_dense) and ((cov is not None) and isinstance(cov,pd.DataFrame)) or ((pre is not None) and isinstance(pre,pd.DataFrame))
    if as_series:
        return pd.Series(index=cov.columns, data=w)
    else:
        return w

    


EQUAL_LONG_PORT = [ equal_long_port ]
EQUAL_LS_PORT = []
EQUAL_PORT = EQUAL_LONG_PORT + EQUAL_LS_PORT


from precise.skaters.covariance.buffactory import _buf
import pandas as pd
import numpy as np

# Thin wrapper for some Riskfolio-Lib methods
# For full flexibility refer to the package https://github.com/dcajasn/Riskfolio-Lib/blob/master/examples


RPL_RISK_METRICS = ['vol', 'MV', 'MAD', 'MSV', 'FLPM', 'SLPM',
       'VaR','CVaR', 'EVaR', 'WR', 'MDD', 'ADD',
       'DaR', 'CDaR', 'EDaR', 'UCI', 'MDD_Rel', 'ADD_Rel',
       'DaR_Rel', 'CDaR_Rel', 'EDaR_Rel', 'UCI_Rel']


def rfl_hrp_factory(s, y, k=1, metric='MV', n_buffer:int=100, e=1):
    """
         Hierarchical Risk Parity
    """
    assert metric in RPL_RISK_METRICS, metric+' is not a valid RPL_RISK_METRIC '
    optim_kwargs = {'rm':metric,'model':'HRP'}
    return rpl_manager_factory(s=s, y=y, k=k, e=e, cls_name='HCPortfolio', n_buffer=n_buffer, optim_kwargs=optim_kwargs)


def rpl_manager_factory(s, y, k=1, e=1, cls_name='HCPortfolio', optim_kwargs:dict=None, n_buffer:int=1000):
    """
         Applies riskfolio-lib routines to buffer
    """
    fun_kwargs = {'rp_optim_kwargs': optim_kwargs}
    fun_kwargs.update({'port_cls_name': cls_name})
    s = _buf(funcs=[_rpl_func], func_names=['weights'], func_kwargs=[fun_kwargs], s=s, x=y,
                e=e, n_buffer=n_buffer)
    w = s.get('weights')
    if w is None:
        w = np.ones(len(y))
    return w, s


def _rpl_func(xs, port_cls_name, rp_optim_kwargs:dict=None):
    import riskfolio as rp
    if rp_optim_kwargs is None:
        rp_optim_kwargs = {}
    returns = pd.DataFrame(data=xs)
    try:
        portfolio_cls = getattr(rp, port_cls_name)
    except:
        raise ValueError(port_cls_name+' is not a valid class name in riskfolio-lib')
    port = portfolio_cls(returns=returns)
    w_dict = port.optimization(**rp_optim_kwargs)
    w = w_dict['weights'].values
    return w








import pathlib
from setuptools import setup, find_packages

HERE = pathlib.Path(__file__).parent

README = (HERE / "README.md").read_text()

setup(
    name="precise",
    version="0.5.5",
    description="Online covariance, precision, portfolios and ensembles",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/microprediction/precise",
    author="microprediction",
    author_email="peter.cotton@microprediction.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
    ],
    packages=["precise","precise.z",
              "precise.skaters",
              "precise.skaters.covariance",
              "precise.skaters.covarianceutil",
               "precise.skaters.scalarutil",
              "precise.skaters.portfoliostatic",
              "precise.skaters.portfolioutil",
              "precise.skaters.location",
              "precise.skaters.managers",
              "precise.skaters.managerutil",
              "precise.skaters.locationutil",
              "precise.skatertools",
              "precise.skatertools.data",
              "precise.skatertools.ensembling",
              "precise.skatertools.m6",
              "precise.skatertools.syntheticdata",
              "precise.skatervaluation",
              "precise.skatervaluation.queues",
              "precise.skatervaluation.battleutil",
              'precise.skatervaluation.battlescripts',
              "precise.skatervaluation.battlelatex",
              'precise.skatervaluation.battlescripts.cov_likelihood',
              'precise.skatervaluation.battlescripts.manager_info',
              'precise.skatervaluation.battlescripts.manager_var'
              ],
    test_suite='pytest',
    tests_require=['pytest'],
    include_package_data=True,
    install_requires=['numpy','momentum>=0.2.5','kmeans1d','runthis','scikit-learn','osqp','latextable','tomark',
                      'pandas_datareader','seriate','pandas','scipy>=1.7.3','pyportfolioopt','riskfolio-lib'],
    entry_points={
        "console_scripts": [
            "precise=precise.__main__:main",
        ]
    },
)



import numpy as np

def normalize(x):
    ### See also portfolioutil.portfunctions.normalize_portfolio
    try:
        return x/sum(x)
    except (TypeError, ArithmeticError):
        try:
            return [xi/sum(x) for xi in x]
        except (TypeError, ArithmeticError):
            if sum(x)<1e-12:
                return x
            else:
                raise ValueError('??')


def scatter(x):
    """
         matrix  x x^t
    """
    x1 = np.atleast_2d(x)
    xt = np.transpose(x1)
    s = np.dot(xt,x1)
    assert np.array_equal( np.shape(s), [len(x),len(x)] )
    return s


if __name__=='__main__':
    from pprint import pprint
    x = np.random.randn(5)
    pprint(scatter(x))


from precise.skaters.managers.weakmanagerfactory import weak_pm_manager_factory, weak_ewa_manager_factory, weak_manager_factory
from precise.skaters.covariance.bufsk import buf_sk_glcv_pcov_d0_n100, buf_sk_glcv_pcov_d0_n100_t0, buf_sk_ld_pcov_d0_n100, buf_sk_mcd_pcov_d0_n100, buf_sk_ld_pcov_d1_n100


def weak_pm_t0_d0_r025_n50_long_manager(y, s, k=1, e=1):
    assert k==1
    return weak_pm_manager_factory(y=y,s=s, target=0, r=0.025, n_emp=50, e=e)


def weak_ewa_t0_d0_r025_n50_long_manager(y, s, k=1, e=1):
    assert k==1
    return weak_ewa_manager_factory(y=y,s=s, r=0.025, n_emp=50, e=e)


def weak_pm_t0_d0_r050_n50_long_manager(y, s, k=1, e=1):
    assert k==1
    return weak_pm_manager_factory(y=y,s=s, target=0, r=0.050, n_emp=50, e=e)


def weak_ewa_t0_d0_r050_n50_long_manager(y, s, k=1, e=1):
    assert k==1
    return weak_ewa_manager_factory(y=y,s=s, r=0.050, n_emp=50, e=e)

# Sklean


def weak_sk_ld_pcov_d0_n100_long_manager(y,s,k=1, e=1):
    return weak_manager_factory(y=y,s=s,f=buf_sk_ld_pcov_d0_n100, e=e)


def weak_sk_glcv_pcov_d0_n100_long_manager(y,s,k=1, e=1):
    return weak_manager_factory(y=y,s=s,f=buf_sk_glcv_pcov_d0_n100, e=e)


def weak_sk_glcv_pcov_d0_n100_t0_long_manager(y,s,k=1, e=1):
    return weak_manager_factory(y=y,s=s,f=buf_sk_glcv_pcov_d0_n100_t0, e=e)


def weak_sk_mcd_pcov_d0_n100_long_manager(y,s,k=1, e=1):
    return weak_manager_factory(y=y,s=s,f=buf_sk_mcd_pcov_d0_n100, e=e)


WEAK_LONG_MANAGERS = [weak_pm_t0_d0_r025_n50_long_manager,
                      weak_ewa_t0_d0_r025_n50_long_manager,
                      weak_pm_t0_d0_r050_n50_long_manager,
                      weak_ewa_t0_d0_r050_n50_long_manager,
                      weak_sk_ld_pcov_d0_n100_long_manager,
                      weak_sk_glcv_pcov_d0_n100_long_manager,
                      weak_sk_glcv_pcov_d0_n100_t0_long_manager,
                      weak_sk_mcd_pcov_d0_n100_long_manager
                      ]


from precise.skatertools.data.equityhistorical import random_cached_equity_dense
import traceback
import os
from uuid import uuid4
import json
from pprint import pprint
from precise.whereami import TESTSERROR
import pathlib


def manager_test_run(mgr,n_obs=50,n_dim=7):
    """
       Test manager and log traceback to /testserrors
    """
    xs = random_cached_equity_dense(k=1, n_obs=n_obs, n_dim=n_dim, as_frame=False)

    try:
        s = {}
        for y in xs[:n_obs - 5]:
            w, s = mgr(y=y, s=s, e=-1)

        for y in xs[-5:]:
            w, s = mgr(y=y, s=s, e=1)
    except Exception as e:
        error_data = {'traceback':traceback.format_exc(),
         'exception':str(e),
         'manager':mgr.__name__}
        manager_error_dir = os.path.join(TESTSERROR,'managers')
        pathlib.Path(manager_error_dir).mkdir(parents=True, exist_ok=True)
        error_file = os.path.join(manager_error_dir,mgr.__name__+'_'+str(uuid4())+'.json')
        pprint(error_data)
        with open(error_file, 'wt') as fh:
            json.dump(error_data,fh)
    return w






import numpy as np
import pandas as pd


def affection_tilt(covdf:pd.DataFrame, love:[str]=None, hate:[str]=None, intensity=1.0):
    """ Convert list of tickers loved and hated into relative means
    :param covdf:   covariance pandas dataframe with tickers as columns and index
    :param love:
    :param hate:
    :return:
    """
    mean_var = np.mean(np.diag(covdf.values))
    tickers = list(covdf.columns)
    n_dim = len(tickers)

    mu = np.zeros(n_dim)
    if love is not None:
        for ticker in love:
            if ticker in tickers:
                mu[tickers.index(ticker)] = intensity * 0.1 * mean_var / 2

    if hate is not None:
        for ticker in hate:
            if ticker in tickers:
                mu[tickers.index(ticker)] = -intensity * 0.1 * mean_var / 2

    return mu


import numpy as np
from precise.skaters.covarianceutil.adjacency import centroid_precision_adjacency
from precise.skaters.covarianceutil.covfunctions import affine_inversion
from precise.skaters.covariance.ewaemp import ewa_emp_pcov_factory

# A factory for approaches that utilize episodically estimated sparsity in precision matrix.
# Yielding a family of autonomous algorithms inspired by a paper by Lee and Zhong

#  Algorithm steps and defaults
#  ----------------------------
#  1. A global cov matrix is maintained  (ewa_pcov)
#  2. A buffer of past observations is maintained
#  3. Each epoch an adjacency matrix for precision is estimated from the global cov (centroid method)
#  4. Local pre models are re-booted, assuming the new structure (ewa_pcov)
#  5. At any time a global pre matrix can be computed from local pre matrices (affine inversion)
#  6. At any time a global cov matrix can be computed by inverting global pre (affine inversion)

# Here are the defaults:

# Step 1
default_f_global = ewa_emp_pcov_factory
DEFAULT_F_GLOBAL_KWARGS = {'r':0.01}

# Step 2
DEFAULT_N_EPOCH = 100

# Step 3
def default_adjacency_func(cov, phi=1.01, lmbd=0.01):
    """ Default way to specify sparsity in precision matrix
    :param cov: (n_dim, n_dim)
    :return: (n_dim, n_dim) bool
    """
    pre = affine_inversion(a=cov, phi=phi, lmbd=lmbd)
    return centroid_precision_adjacency(pre=pre)
DEFAULT_ADJ_FUNC_KWARGS = {'phi':1.01,'lmbd':0.01}


# Step 4
default_f_local = ewa_emp_pcov_factory
DEFAULT_F_LOCAL_KWARGS = {'r':0.01}

# Step 5
def default_local_pre_func(cov, phi, lmbd):
    """ Default way to invert local cov matrices
    """
    return affine_inversion(a=cov, phi=phi, lmbd=lmbd)
DEFAULT_LOCAL_PRE_FUNC_KWARGS = {'phi':1.1,'lmbd':0.1}

# Step 6
default_global_cov_func = default_local_pre_func
DEFAULT_GLOBAL_COV_FUNC_KWARGS = {'phi':1.01,'lmbd':0.1}


def ewa_emp_lz_pcov_factory_rg_rl_n(y, s, rg, rl, n, e):
    """
    :param y:
    :param s:
    :param n_epoch:     Length of epoch
    :param rg:
    :param rl:
    :param e:
    :return:
    """
    f_global_kwargs = {'r':rg}
    f_local_kwargs = {'r':rl}
    return lz_factory(y=y,s=s,n_epoch=n,
                      f_local =  ewa_emp_pcov_factory,
                      f_global = ewa_emp_pcov_factory,
                      f_global_kwargs=f_global_kwargs,
                      f_local_kwargs=f_local_kwargs,
                      e=e)


def lz_factory(y, s:dict, n_epoch=DEFAULT_N_EPOCH,
               f_global=None, f_global_kwargs=None,
               f_local=None, f_local_kwargs=None,
               e=1,
               adj_func=None, adj_func_kwargs:dict=None,
               local_pre_func=None, local_pre_func_kwargs:dict=None,
               global_cov_func=None, global_cov_func_kwargs:dict=None):
    """

      Le and Zhong factory

    :param y:                Vector of observations
    :param s:                State
    :param n_epoch:          Length of epoch during which adjacency matrix is held constant
    :param f_global:         Skater maintaining a global estimate of covariance
    :param f_local:          Skater used to maintain n_dim local estimates, one for each variable
    :param e:                Set <0 to skip global cov, pre calculations
    :param adj_func:         Takes pre matrix and returns binary of same shape (where non-zeros should go)
    :param local_pre_func:   Function to take local cov matrices and produce precision matrices
    :param global_cov_func:  Function to take global precision matrix and produce global cov matrix
    :return: x, x_cov, s
    """

    # Mildly tedious stuff...
    if adj_func is None:
        adj_func = default_adjacency_func
    if local_pre_func is None:
        local_pre_func = default_local_pre_func
    if global_cov_func is None:
        global_cov_func = default_global_cov_func
    if adj_func_kwargs is None:
        adj_func_kwargs = DEFAULT_ADJ_FUNC_KWARGS
    if local_pre_func_kwargs is None:
        local_pre_func_kwargs = DEFAULT_LOCAL_PRE_FUNC_KWARGS
    if global_cov_func_kwargs is None:
        global_cov_func_kwargs = DEFAULT_GLOBAL_COV_FUNC_KWARGS

    if f_global_kwargs is None:
        f_global_kwargs = {}
    if f_local_kwargs is None:
        f_local_kwargs = {}

    # Initialization of global state s['g'] and local states s['l']['states']
    n_dim = len(y)
    if not s:
        s = {'g':{},          # Global state
             'l':{'states':[{} for _ in range(n_dim)]}, # Local states
             'buffer':[],     # Observation buffer
             'count':0,
             'warm':0,        # Local models are primed
             'stale':0}       # Global pre/cov is stale

    # Maintain a global covariance model
    s['count'] += 1
    x_gl, gl_cov, s['g'] = f_global(s=s['g'], y=y, **f_global_kwargs)
    if s['count'] % n_epoch == 0:
        # Trigger adjacency matrix discovery from the global model
        s['l']['adj'] = adj_func(gl_cov, **adj_func_kwargs)

        # Prepare the "B" matrices in the paper
        s['l']['n_dims'] = [int(s) for s in np.sum(s['l']['adj'], axis=0)]
        s['l']['bs'] = dict()
        for i, nd in enumerate(s['l']['n_dims']):
            b_ = np.zeros(shape=(n_dim, nd))
            ndxs = s['l']['adj'][:, i]
            b_[ndxs, :] = np.eye(nd)
            s['l']['bs'][i] = b_

        # Reset local states ...
        s['l']['states'] = [{} for _ in range(n_dim)]
        #  ... but replay data from the buffer through local states once
        for i in range(n_dim):
            ls = {}
            indx = s['l']['adj'][:, i]
            for yt in s['buffer']:
                yti = yt[indx]
                xl, xl_cov, ls = f_local(s=ls,y=yti, **f_local_kwargs )
            s['l']['states'][i] = ls

        # Blow away the buffer and declare it ready
        s['buffer'] = []
        s['warm'] = 1

    # Add current data vector to buffer
    s['buffer'].append(y)
    if len(s['buffer'])>n_epoch:
        s['buffer'].pop(0)

    # If local models are initialized, process one data point
    if s['warm']:
        # Update local states and most recent cov estimates
        s['l']['covs'] = [ None for _ in range(n_dim) ]
        for i in range(n_dim):
            indx = s['l']['adj'][:, i]
            yi = y[indx]
            sl = s['l']['states'][i]
            _, s['l']['covs'][i], sl = f_local(s=sl, y=yi, **f_local_kwargs)
            s['l']['states'][i] = sl



    # If allowed time, compute precision and covariance
    #  - User can set e<0 to skip this step)
    #  - User can supply a "fake" global_cov_func to avoid that overhead, if they want
    if s['warm']:
        if (e>=0):
                # Reconstruct precision - see paper for explanation
                omega = np.eye(n_dim)
                for i,covi in enumerate(s['l']['covs']):
                    prei = local_pre_func(covi, **local_pre_func_kwargs)
                    ei = np.zeros(shape=(n_dim,1))
                    ei[i] = 1.0
                    bi = s['l']['bs'][i]
                    fi = np.matmul(bi.T,ei)
                    Sfi = np.matmul(prei,fi)
                    wi = np.matmul(bi,Sfi)
                    omega[:,i] = wi[:,0]
                s['pre'] = omega
                s['cov'] = global_cov_func(omega,**global_cov_func_kwargs)
                s['stale']=0
        else:
            s['stale']=1

    # Return cov is fresh, else fall back to global model
    if (not s['warm']) or s['stale']:
        return x_gl, gl_cov, s
    else:
        return x_gl, s['cov'], s
















import pandas as pd
import numpy as np
import random


def random_cached_equity_dense(k, n_obs, n_dim=10, as_frame=False, **ignore):
    """
         Log price differences without missing values
         :param k - days

    """
    pd.options.mode.chained_assignment = None
    df = _sparse_cached_equity(k=k)
    n_samples = len(df.index)
    assert n_samples>n_obs
    n_start = random.choice(range(n_samples-n_obs))
    n_end = n_start + n_obs
    df_sub = df[n_start:n_end]
    df_sub.dropna(how='any', inplace=True, axis=1)
    df_sub.drop(df.columns[0], inplace=True, axis=1)
    size = min(len(df_sub.columns),n_dim)
    selected = np.random.choice(df_sub.columns, size=size, replace=False)
    df_some = df_sub[selected]
    return df_some if as_frame else df_some.values


def _sparse_cached_equity(k=1):
    """ Retrieve daily or less frequent price changes
    :param k:  difference measured in business days
    :returns   pd.DataFrame with NaNs
    """
    CSV ='https://raw.githubusercontent.com/microprediction/precisedata/main/stocks/log_price_diff_K_part_N.csv'
    dfs = [ pd.read_csv(CSV.replace('K',str(k)).replace('N',str(part)) ) for part in list(range(1,4)) ]
    df_merged = pd.concat(dfs,axis=1)
    return df_merged



if __name__=='__main__':
    df = _sparse_cached_equity(k=1)
    print(df[:3])
    dg = random_cached_equity_dense(k=1, n_obs=50, as_frame=True)
    print(dg)


from precise.skaters.portfoliostatic.diagportfactory import diagonal_portfolio_factory
from precise.skaters.portfoliostatic.diagalloc import diag_alloc
from pprint import pprint
from functools import partial
from precise.skaters.covarianceutil.covfunctions import schur_complement, \
    to_symmetric, multiply_by_inverse, inverse_multiply, is_positive_def, nearest_pos_def
from scipy.optimize import root_scalar
from precise.skaters.portfoliostatic.schurportutil import symmetric_step_up_matrix, even_split
from precise.skaters.covarianceutil.covfunctions import try_invert, cov_distance
from seriate import seriate
import numpy as np
from precise.skaters.portfoliostatic.equalport import equal_long_port


def schur_portfolio_factory(seriator=None, alloc=None, port=None, splitter=None, cov=None, pre=None, n_split=5, gamma=1.0, delta=0.0):
    """
          A new top-down method
    """
    port_kwargs = {'gamma':gamma, 'delta':delta}
    return hierarchical_seriation_portfolio_factory(seriator=seriator, alloc=alloc, port=port, splitter=splitter, cov=cov, pre=pre, n_split=n_split, port_kwargs=port_kwargs)


def hierarchical_seriation_portfolio_factory(seriator=None, alloc=None, port=None, splitter=None, cov=None, pre=None, n_split=5, port_kwargs=None):
    """
        A class of algorithms that apply seriation ordering then allocate in top-down fashion

    :param alloc:      Decides how much capital to split between portfolios  [covs] -> [ float ]
    :param port:       Computes a portfolio     cov -> [ float ]    (Used on the leaves only)
    :param splitter:   Splits into two groups   cov -> (n1,n2)
    :param gamma:      (0,1) How far to move towards Schur complement
    :return:
    """
    # Remark. The port and alloc need not be cut from the same cloth

    if alloc is None:
        alloc = diag_alloc

    if port is None:
        port = diagonal_portfolio_factory

    if splitter is None:
        splitter = partial( even_split, n_split=n_split )

    if port_kwargs is None:
        port_kwargs = {}

    port_kwargs.update({'splitter':splitter,'alloc':alloc,'port':port})
    return corr_seriation_portfolio_factory(seriator=seriator, port=hierarchical_seriated_portfolio_factory, port_kwargs=port_kwargs, cov=cov, pre=pre)


def corr_seriation_portfolio_factory(port, port_kwargs:dict=None, seriator=None, cov=None, pre=None)->np.ndarray:
    """
        A class of methods whose first step is seriation using correlation

    :param seriator:          Takes a distance matrix and returns an ordering
    :param port:              Portfolio generator
    :param port_kwargs:       Arguments to portfolio generator, other than 'cov' and/or 'pre'
    :param cov:               Original portfolio in arbitrary order
    :param pre:               Original precision matrix in arbitrary order
    :return: w                Portfolio weights in original ordering
    """
    if cov is None:
        cov = try_invert(pre)

    if seriator is None:
        seriator = seriate

    if port_kwargs is None:
        port_kwargs = {}

    if any(np.diag(cov)<1e-6):
        return equal_long_port(cov=cov)
    else:
        # Establish ordering using seriator and corr distances
        try:
            cov_dist = cov_distance(cov)
            ndx = seriator(cov_dist)
            inv_ndx = np.argsort(ndx)
            cov_cols = cov[:,ndx]
            cov_back = cov_cols[:,inv_ndx]
            assert np.allclose(cov,cov_back)
            ordered_cov = cov_cols[ndx,:]
        except Exception as e:
            print('warning: Seriation failed ')
            return equal_long_port(cov=cov)

        # Allocate capital to ordered assets
        ordered_w = port(cov=ordered_cov, **port_kwargs)

        # Return to original ordering
        try:
            w = ordered_w[inv_ndx]
        except TypeError:
            print('Warning: '+port.__name__+' returns list not array - should really fix this ')
            w = np.array(ordered_w)[inv_ndx]
        return np.array(w)


def hierarchical_seriated_portfolio_factory(alloc, cov, port, splitter, gamma:float=0.0, delta:float=0):
    """
        Assumes assets have been ordered already
    """
    n1, n2 = splitter(cov)
    if n1==0 or n2==0:
        w = port(cov)
        if isinstance(w,list):
            print('Warning: '+port.__name__+' returns list not array ')
            w = np.array(w)
        return w
    else:
        if abs(gamma)<1e-6:
            # Hierarchical risk parity (Lopez de Prado)
            w = hierarchical_risk_parity(cov=cov, n1=n1, port=port, alloc=alloc, splitter=splitter)
        else:
            # Schur complementary portfolio construction (yours truly)
            w = hierarchical_schur_complementary_portfolio(cov=cov, n1=n1, port=port, alloc=alloc, splitter=splitter, gamma=gamma, delta=delta)
        return w


def hierarchical_schur_complementary_portfolio(cov, n1, port, alloc, splitter, delta=0.0, gamma=0.5):
    """
        An experimental way to split allocation
    """
    A = cov[:n1, :n1]
    D = cov[n1:, n1:]
    B = cov[:n1, n1:]
    C = cov[n1:, :n1]  #  = B.T

    if delta>0:
        # Haven't tried this yet :)
        rhoB = np.mean(B,axis=None)
        rhoCov_raw = cov - rhoB*np.ones_like(cov)
        rhoCov = nearest_pos_def(rhoCov_raw)
        return hierarchical_schur_complementary_portfolio(cov=rhoCov, n1=n1, port=port, alloc=alloc, splitter=splitter, gamma=gamma)
    else:
        if gamma>0.0:
            # Augment the cov matrices before passing down
            max_gamma = _maximal_gamma(A=A, B=B, C=C, D=D)
            augA = pseudo_schur_complement(A=A, B=B, C=C, D=D, gamma=gamma * max_gamma)
            augD = pseudo_schur_complement(A=D, B=C, C=B, D=A, gamma=gamma * max_gamma)
            if not is_positive_def(augA):
                Ag = nearest_pos_def(augA)
            else:
                Ag = augA
            if not is_positive_def(augD):
                Dg = nearest_pos_def(augD)
            else:
                Dg = augD
            reductionD = np.linalg.norm(Dg)/np.linalg.norm(D)
            reductionA = np.linalg.norm(Ag)/np.linalg.norm(A)
            reductionRatioA = reductionA/reductionD
        else:
            reductionRatioA = 1.0
            Ag = A
            Dg = D
        wA = hierarchical_seriated_portfolio_factory(alloc=alloc, cov=Ag, port=port, splitter=splitter, gamma=gamma)
        wD = hierarchical_seriated_portfolio_factory(alloc=alloc, cov=Dg, port=port, splitter=splitter, gamma=gamma)
        aA, aD = alloc(covs=[Ag, Dg])
        aA_original, aD_original = alloc( covs=[A,D])
        allocationRatioA = (aA/aA_original)
        if False:
            info = {'reductionA':reductionA,
                    'reductionD':reductionD,
                    'reductionRatioA':reductionRatioA,
                    'allocationRatioA':allocationRatioA}
            pprint(info)
        w = np.concatenate([aA * np.array(wA), aD * np.array(wD)])
        return np.array(w)



def hierarchical_risk_parity(cov, n1, port, alloc, splitter):
    """
         Recursive hierarchical risk parity
    """
    # https://papers.ssrn.com/sol3/papers.cfm?abstract_id=2708678
    A = cov[:n1, :n1]
    D = cov[n1:, n1:]
    wA = hierarchical_seriated_portfolio_factory(alloc=alloc, cov=A, port=port, splitter=splitter, gamma=0.0)
    wD = hierarchical_seriated_portfolio_factory(alloc=alloc, cov=D, port=port, splitter=splitter, gamma=0.0)
    aA, aD = alloc(covs=[A, D])
    w = np.concatenate([aA * np.array(wA), aD * np.array(wD)])
    return w



def pseudo_schur_complement(A, B, C, D, gamma, warn=False):
    """
       Augmented cov matrix for "A" inspired by the Schur complement
    """
    try:
        Ac_raw = schur_complement(A=A, B=B, C=C, D=D, gamma=gamma)
        nA = np.shape(A)[0]
        nD = np.shape(D)[0]
        Ac = to_symmetric(Ac_raw)
        M = symmetric_step_up_matrix(n1=nA, n2=nD)
        Mt = np.transpose(M)
        BDinv = multiply_by_inverse(B, D, throw=False)
        BDinvMt = np.dot(BDinv, Mt)
        Ra = np.eye(nA) - gamma * BDinvMt
        Ag = inverse_multiply(Ra, Ac, throw=False, warn=False)
    except np.linalg.LinAlgError:
        if warn:
            print('Pseudo-schur failed, falling back to A')
        Ag = A
    return Ag


def _maximal_gamma(A,B,C,D):

    def _gamma_objective(gamma, A, B, C, D):
        Ag = pseudo_schur_complement(A=A, B=B, C=C, D=D, gamma=gamma)
        Dg = pseudo_schur_complement(A=D, B=C, C=B, D=A, gamma=gamma)
        pos_def = is_positive_def(Ag) and is_positive_def(Dg)
        return -0.01 if pos_def else 1.0

    try:
        sol = root_scalar(f=_gamma_objective, args=(A,B,C,D), method='bisect', x0=0.25,
                          x1=0.5, xtol=0.05, bracket=(0,0.95), maxiter=5)
        return min(max(sol.root - 0.1, 0), 1.0)
    except ValueError:
        return 0.0



if __name__=='__main__':
    M = symmetric_step_up_matrix(n1=7, n2=6)
    print(np.dot(M, np.ones(6)))
    print(np.dot(M.transpose()*6/7,np.ones(7)))


import numpy as np
from precise.skatertools.data.equitylive import get_prices
import pandas as pd
from pprint import pprint
from precise.skaters.covariance.runemp import run_emp_pcov_d0


def m6_data(interval='d', n_dim=100, n_obs=300):
    constituents = pd.read_csv(
        'https://raw.githubusercontent.com/microprediction/m6/main/data/official/M6_Universe.csv')[:n_dim]
    tickers = constituents['symbol'].values
    if (interval=='m') and (n_obs>60):
        print('Too many obs, switching to daily ')
        interval = 'd'
    df = pd.DataFrame(columns=tickers)
    for ticker in tickers:
        closing_prices = get_prices(ticker=ticker, n_obs=n_obs+1, interval=interval)
        while len(closing_prices) < n_obs+1:
            closing_prices = list(closing_prices) + list(closing_prices)
        closing_prices = closing_prices[-n_obs:]
        df[ticker] = np.diff(np.log(closing_prices))
    return df


def m6_cov(f=None, interval='d', n_dim=100, n_obs=300):
    """ Use any skater to estimate daily (by default) covariance
    :param f: cov skater
    :return:
    """
    if f is None:
        f = run_emp_pcov_d0
    df = m6_data(interval=interval, n_dim=n_dim, n_obs=n_obs)
    tickers = list(df.columns)

    s = {}
    for y in df.values:
        x_mean, x_cov, s = f(s=s, y=y)
    return pd.DataFrame(index=tickers, columns=tickers, data=x_cov)


def m6_corr(f=None, n_dim=100, interval='d',n_obs=300):
    from precise.skaters.covarianceutil.datafunctions import cov_to_corrcoef
    covdf = m6_cov(f=f, n_dim=n_dim, interval=interval, n_obs=n_obs)
    tickers = list(covdf.columns)
    corr = cov_to_corrcoef(covdf.values)
    dfc = pd.DataFrame(index=tickers, columns=tickers, data=corr)
    return dfc


if __name__=='__main__':
    from precise.skaters.covariance.ewaemp import ewa_emp_pcov_d0_r01 as f
    pprint(m6_corr(interval='d',n_dim=4,f=f))



import os
from pathlib import Path
from pprint import pprint
TOP = os.path.dirname(os.path.abspath(__file__))
BATTLE_RESULTS_DIR = os.path.join(TOP, 'skatervaluation', 'battleresults')
GITHUB_COV_SKATERS = 'https://github.com/microprediction/precise/blob/main/precise/skaters/covariance/'
GITHUB_MANAGERS = 'https://github.com/microprediction/precise/blob/main/precise/skaters/managers/'
ROOT = Path(TOP).parent.absolute()
M6_EXAMPLES = os.path.join(ROOT,'examples_m6')
MANAGER_MANIFEST = os.path.join(ROOT,'LISTING_OF_MANAGERS.md')
COV_SKATER_MANIFEST = os.path.join(ROOT,'LISTING_OF_COV_SKATERS.md')
TESTSERROR = os.path.join(ROOT,'testserrors')


def url_from_skater_name(name:str)->str:
    """
         Infer the location in GitHub for the skater's code
         e.g.  buf_emp_pcov_d0_n100 ->https://github.com/microprediction/precise/blob/main/precise/skaters/covariance/bufemp.py
    """
    return GITHUB_COV_SKATERS + ''.join(name.split('_')[:2])+'.py'


def url_from_manager_name(name:str)->str:
    return GITHUB_MANAGERS + ''.join(name.split('_')[:1])+'managers.py'



if __name__=='__main__':
    print(TOP)
    print(url_from_skater_name(name='buf_emp_pcov_d0_n100'))
    print(M6_EXAMPLES)



from sklearn.datasets import make_spd_matrix
import numpy as np
import random


def create_factor_dataset(n, n_dim):
    Eidio = make_spd_matrix(n_dim=n_dim)
    Xidio = np.random.multivariate_normal(mean=np.zeros(n_dim), cov = Eidio, size=n)
    D = np.diag([ random.choice([0.25,1.0,4]) for _ in range(n_dim) ])
    Ecommon = 0.4*np.eye(n_dim) + 0.6*np.ones(n_dim)
    cov_common = np.matmul(np.matmul(D,Ecommon),D)
    Xcommon = np.random.multivariate_normal(mean=np.zeros(n_dim), cov = Ecommon, size=n)
    x = Xcommon + Xidio
    return x



def create_disjoint_factor_dataset(n, n_dims:[int]):
    xs = [ create_factor_dataset(n,n_dim) for n_dim in n_dims ]
    x = xs[0]
    for xi in xs[1:]:
        x = np.concatenate( [x,xi], axis=1)
    return x



def create_band_dataset(n, n_dim, n_bands=5):
    z1 = create_factor_dataset(n, n_dim)
    z2 = create_factor_dataset(n, n_dim)
    R = np.zeros(shape=(n_dim,n_dim))
    for i in range(n_dim):
        R[i,i] = 0.75
        try:
            for j in range(1,n_bands+1):
                R[i,i+j] = np.random.rand()*0.5
        except IndexError:
            pass
    x = np.matmul(z1,R.T) + z1
    return x






from precise.skatervaluation.battleutil.compilingeloratings import elo_from_win_files
from texttable import Texttable
from latextable import draw_latex


def leaderboard_to_latex(leaderboard, caption, label):
     MANAGER_REPLACEMENTS = {'_t0':' target=0',
                     '_pm':' partial moments  ',
                     '_emp':' empirical ',
                     '_ppo':' ',
                     '_r':' r=0.',
                     '_g050':' \gamma=0.5',
                     '_g100': ' \gamma=1',
                     '_n':' window=',
                     '_long_manager':'',
                     '_vol':' (min-vol)',
                    '_quad':' (quadratic)',
                    '_sharpe':' (Sharpe)',
                    '_weak':' (weak)',
                    '_schur':' (Schur)',
                    '_hrp':' (HRP)',
                    '_pcov':'',
                    '_ld':' Ledoit Wolf',
                    '_glcv':' Graphical lasso w/ cv',
                    '_gl':' Graphical lasso',
                    '_mcd':' min-cov-det',
                    '_d0':'',
                    '_ewa':' expon weighted ',
                    '_d1':'',
                    '_sk':'',
                    'hrp':'HRP',
                    'schur':'Schur',
                    'ppo':''}

     tbl = Texttable()
     tbl.set_cols_align(["l", "l", "c"])
     tbl.set_cols_valign(["t", "m", "b"])
     rows = [['Elo', 'Approach', 'CPU']]
     for name, res in leaderboard.items():
          for _ in range(2):
               for k,v in MANAGER_REPLACEMENTS.items():
                    name = name.replace(k,v)
          name = name.replace('_',' ')
          if isinstance(res,tuple):
               elo_rating = round(res[0],ndigits=0)
               cpu = round(res[1],ndigits=0) if res[1] is not None else 'N/A'
          else:
               elo_rating = round(res,ndigits=0)
               cpu = 'N/A'
          rows.append([elo_rating,name,cpu])
     tbl.add_rows(rows=rows)
     return draw_latex(tbl, caption=caption, label=label)


def elo_latex_table(genre, category):
     elo_results = dict( elo_from_win_files(genre=genre) )
     try:
          leaderboard = elo_results[category]
     except KeyError:
          print(category+' is not in '+genre)
          print('Try one of '+str(list(elo_results.keys())))
          raise KeyError

     CATEGORY_REPLACEMENTS = {'_p':' with asset count ',
                              '_n':' using historical data length '}

     description = category
     for k,v in CATEGORY_REPLACEMENTS.items():
          description = description.replace(k,v)
     description = genre.replace('_',' ') + ' for ' + description.replace('_',' ') + '.'
     description = description[0].upper()+description[1:]
     label='tab:'+genre+'_'+category
     return leaderboard_to_latex(leaderboard=leaderboard, caption=description, label=label )


if __name__=='__main__':
     ltx = elo_latex_table(genre='manager_var',category='stocks_1_days_p200_n200')
     print(ltx)









import pandas as pd
import numpy as np


def square_to_square_dataframe(df, cov_func, **cov_func_kwargs):
    """ Map square dataframe to square dataframe """
    return pd.DataFrame( index=df.index, columns=df.columns, data=cov_func(df.values, **cov_func_kwargs))


def square_to_column_series(df, cov_func, **cov_func_kwargs):
    """ Map square dataframe to Series indexed by columns """
    return pd.Series( index=df.columns, data=cov_func(df.values, **cov_func_kwargs) )


def square_to_index_series(df, cov_func, **cov_func_kwargs):
    """ Map square dataframe to Series indexed by index """
    return pd.Series( index=df.index, data=cov_func(df.values, **cov_func_kwargs) )


def data_to_square_dataframe(df, data_func, **data_func_kwargs):
    return pd.DataFrame( index=df.columns, columns=df.columns, data=data_func(df.values, **data_func_kwargs))




def dict_or_series_values(d):
    # Remark: https://stackoverflow.com/questions/835092/python-dictionary-are-keys-and-values-always-the-same-order
    if isinstance(d,dict):
        return np.array(list(d.values()))
    elif isinstance(d,pd.Series):
        return d.values


def dict_or_series_to_scalar(d, func, **dict_func_kwargs):
    x = dict_or_series_values(d)
    return dict(zip(d.keys(), func(x, **dict_func_kwargs)))


def _square_and_vector_values(a,w):
    """ Takes square and vector arguments, which might be dict/Series/DataFrame
        Assumes columns of a correspond to index of w
    :param a:
    :param w:
    :return: (a,b), (b,)
    """
    if isinstance(a,pd.DataFrame) and isinstance(w, (dict, pd.Series)):
        if isinstance(w, pd.Series):
            w = w.to_dict()
        assert set(w.values()) == set(a.columns)
        w_values = np.array([w[ky] for ky in list(a.columns)])
        a_values = a.values
    elif isinstance(a,pd.DataFrame) and isinstance( w, (dict, pd.Series)):
        raise ValueError('too dangerous')
    elif isinstance(a, pd.DataFrame) and not isinstance( w, (dict, pd.Series)):
        a_values = a.values
        w_values = np.array(w)
    elif ~isinstance(a, pd.DataFrame) and isinstance( w, (dict, pd.Series)):
        a_values = a
        try:
            w_values = w.values()
        except TypeError:
            w_values = w.values
    else:
        raise NotImplementedError()
    return a_values, w_values


def square_and_vector_to_scalar(a, w, func, **func_kwargs):
    """
    :param a:   np 2d array or square dataframe
    :param w:   1d array, list, dict or Series
    :param func:    (a,w, **func_kwards) -> list or 1d array
    :param func_kwargs:
    :return:
    """
    a_values, w_values = _square_and_vector_values(a=a,w=w)
    return func(a_values, w_values, **func_kwargs)


def square_and_vector_to_vector(a, w, func, **func_kwargs):
    """
    :param a:   np 2d array or square dataframe
    :param w:   1d array, list, dict or Series
    :param func:    (a,w, **func_kwards) -> list or 1d array
    :param func_kwargs:
    :return:
    """
    a_values, w_values = _square_and_vector_values(a=a, w=w)
    y_values = func( a_values, w_values, **func_kwargs)
    if isinstance(w, dict):
        return dict(zip( w.keys(), y_values))
    elif isinstance(w, pd.Series):
        return pd.Series(index=w.index, data=y_values)
    else:
        return np.array(y_values)


def vector_to_vector(w, func, **func_kwargs):
    if isinstance(w, dict):
        y_values = func( w.values(), **func_kwargs )
        return dict(zip(w.keys(),y_values))
    elif isinstance(w, pd.Series):
        y_values = func( w.values, **func_kwargs )
        return pd.Series(index=w.index, data=y_values)
    else:
        return func( np.array(w) )








import numpy as np
from precise.skaters.covarianceutil.conventions import X_TYPE, X_DATA_TYPE, is_data
from precise.skaters.covariance.buffactory import buf_pcov_factory
from precise.skaters.covarianceutil.datafunctions import data_population_covariance
import warnings


def buf_sk_factory(cls, y:X_TYPE=None, s:dict=None,  n_buffer:int=100, n_emp=5, cls_kwargs:dict=None, fit_kwargs:dict=None, e=1):
    """
        :param cls  an sklearn covariance estimating class
        :param n_emp:   If we don't yet have n_emp data points, will revert to empirical
    """

    if cls_kwargs is None:
        cls_kwargs = {}

    if fit_kwargs is None:
        fit_kwargs = {}

    def sk_apply_cov(cls, xs, cls_kwargs:dict, fit_kwargs:dict, n_emp:int):

        if len(xs) < n_emp:
            try:
                _pcov = data_population_covariance(xs)
            except:
                _pcov = np.eye(len(xs))
            try:
                _location = np.nanmean(xs,axis=0)
            except:
                _location = np.zeros(len(xs))
            outputs = {'loc':_location,'pcov':_pcov}
        else:
            obj = cls(**cls_kwargs)
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                obj.fit(np.array(xs),**fit_kwargs)
            try:
                outputs = {'loc': obj.location_, 'pcov': obj.covariance_}
            except AttributeError:
                print('obj lacks ._location or ._covariance')
                raise
        return outputs

    func = lambda xs: sk_apply_cov(cls=cls, xs=xs, fit_kwargs=fit_kwargs, n_emp=n_emp, cls_kwargs=cls_kwargs)
    return buf_pcov_factory( func=func, y=y, s=s, n_buffer=n_buffer, e=e)


from precise.skaters.managers.managerfactory import static_cov_manager_factory_d0
from precise.skaters.covariance.ewapm import ewa_pm_factory, ewa_pm_emp_scov_r01_n100
from precise.skaters.portfoliostatic.weakportfactory import weak_portfolio_factory
from precise.skaters.covariance.ewaempfactory import ewa_emp_pcov_factory
from precise.skaters.covariance.bufsk import buf_sk_glcv_pcov_d0_n100, buf_sk_ld_pcov_d0_n100
from functools import partial


def weak_pm_manager_factory(y, s, target, n_emp,r, e=1, a=1.0, b=None):
    """
       Weak portfolio construction using partial moments cov estimation
    """
    f = partial( ewa_pm_factory, k=1,r=r,target=target, n_emp=n_emp )
    port = partial( weak_portfolio_factory, a=a, b=b )
    return static_cov_manager_factory_d0(f=f, port=port, y=y, e=e, s=s)


def weak_ewa_manager_factory(y, s, n_emp, r , e=1, a=1.0, b=None):
    """
       Weak portfolio construction using partial moments cov estimation
    """
    f = partial( ewa_emp_pcov_factory, k=1,r=r, n_emp=n_emp )
    port = partial( weak_portfolio_factory, a=a, b=b )
    return static_cov_manager_factory_d0(f=f, port=port, y=y, e=e, s=s)


def weak_manager_factory(y, s, f, a=1.0, e=1, b=None):
    """
       Weak portfolio construction using any f
    """
    port = partial( weak_portfolio_factory, a=a, b=b )
    return static_cov_manager_factory_d0(f=f, port=port, y=y, s=s, e=e)



from pprint import pprint
from precise.skaters.covarianceutil.covfunctions import affine_shrink, nearest_pos_def
from precise.skaters.portfoliostatic.allstaticport import random_port, PORT
from precise.skaters.covariance.allcovskaters import random_cov_skater, cov_skater_manifest
from precise.skatertools.m6.quintileprobabilities import m6_probabilities
from precise.skatertools.m6.tilting import affection_tilt

# Demonstrates the creation of an entry in the M6 contest
#
#   1. Pick a cov estimator (i.e. a "cov skater"), if you wish
#   2. Pick a portfolio generator, if you wish
#   3. Pick extra shrinkage params, if you wish
#   4. Pick love and hate ticker lists, if you wish


def m6_dump(df,file_name):
    """ Write to CSV with 'ID' in top corner """
    df.columns = ['Rank'+str(i) for i in range(1,6)] + ['Decision']
    df.reset_index().rename(columns={'index': 'ID'}).to_csv(file_name, index=False)


def m6_competition_entry(interval='d', f=None, port=None, n_dim=100, n_samples=5000, n_obs=200,
                         extra_shrink=True,
                                phi=1.1, lmbd=0.03,
                         love:[str]=None,   # List of companies you love
                         hate:[str]=None,   # List of companies you hate
                         intensity:float=1.0,
                         verbose=True):
    """
           Example of generating an M6 Entry
           pip install PyPortfolioOpt

           interval - sampling interval to use for cov estimation
           n_obs    - number of time points to use in cov estimation (max 60 if interval='m')
           n_dim    - Set at 100 for actual contest, but lower to test
           port     - A portfolio creator (see /portfolioutil )
           f        - A cov skater
           extra_shrink - If True, will perform additional shrinkage over and above the skater or portfolio method using
               phi      - (Additional) Ridge parameter, suggest (1,1.5)
               lmbd     - (Additional) Shrinkage parameter, suggest (0,0.5)
           love     - List of tickers of companies whose means will be adjusted upward slightly
           hate     - List of tickers of companies whose means will be adjusted downward slightly

    """
    if love is None:
        love = []

    if hate is None:
        hate = []

    if port is None:
        port = random_port()
        if verbose:
            print('Choosing a cov estimator from the following list ')
            pprint([p.__name__ for p in PORT])
        print('Chose '+port.__name__)


    if f is None:
        print('Choosing a cov estimator from the following list ')
        pprint(cov_skater_manifest())
        f = random_cov_skater()
        print('Choose '+f.__name__)

    print('Computing rank probabilities')
    df_prob, df_cov = m6_probabilities(f=f, interval=interval, n_dim=n_dim,
                                            n_samples=n_samples,
                                            n_obs=n_obs, verbose=verbose,
                                            love=love, hate=hate, intensity=intensity)
    cov = df_cov.values
    if extra_shrink:
        cov = affine_shrink(cov, phi=phi, lmbd=lmbd)
        cov = nearest_pos_def(cov)

    print('Computing portfolio w/ ad-hoc tilt')
    w = port(cov=cov)
    mu = affection_tilt(covdf=df_cov, love=love, hate=hate, intensity=intensity)
    w[mu>0] = w[mu>0] + 1.0/5*len(love)
    w[mu<0] = w[mu<0] - 1.0/2*len(love)

    sum_abs = sum( [abs(wi) for wi in w] )
    w_normalized = [ wi/sum_abs for wi in w]
    w_rounded = [round(wi, 5) for wi in w_normalized]

    entry = df_prob.copy()
    entry['Decision'] = w_rounded
    entry.rename(inplace=True, columns={'0': 'Rank1', '1': 'Rank2', '2': 'Rank3', '3': 'Rank4', '4': 'Rank5'})
    return entry



if __name__=='__main__':
    df = m6_competition_entry()
    m6_dump(df=df,file_name='m6_competition_entry.csv')








import numpy as np
from precise.skaters.covarianceutil.conventions import Y_DATA_TYPE
from precise.skaters.covarianceutil.differencing import d1_factory
from precise.skaters.covariance.bufempfactory import buf_emp_pcov_d0_factory

# Simple, inefficient skaters that use operations on a buffer of past values


def buf_emp_pcov_d0_n20(y:Y_DATA_TYPE, s:dict, k=1, e=1):
    assert k==1
    return buf_emp_pcov_d0_factory(y=y, s=s, n_buffer = 20, e=e)


def buf_emp_pcov_d0_n50(y:Y_DATA_TYPE, s:dict, k=1, e=1):
    assert k==1
    return buf_emp_pcov_d0_factory(y=y, s=s, n_buffer = 50, e=e)


def buf_emp_pcov_d0_n100(y:Y_DATA_TYPE, s:dict, k=1, e=1):
    return buf_emp_pcov_d0_factory(y=y, s=s, n_buffer = 100, e=e)


def buf_emp_pcov_d1_n20(y:Y_DATA_TYPE, s:dict, k=1, e=1):
    assert k==1
    return buf_emp_pcov_d1(y=y, s=s, n_buffer = 20, e=e)


def buf_emp_pcov_d1_n100(y:Y_DATA_TYPE, s:dict, k=1, e=1):
    assert k==1
    return buf_emp_pcov_d1(y=y, s=s, n_buffer = 100, e=e)


def buf_emp_pcov_d1(y:Y_DATA_TYPE, s:dict, n_buffer:int=100, k=1, e=1):
    """
        For when changes are iid gaussian
    """
    assert k==1
    return d1_factory(f = buf_emp_pcov_d0_factory, y=y, s=s, n_buffer=n_buffer, e=e)


def buf_emp_pcov_d1_long_form(y:Y_DATA_TYPE, s:dict, n_buffer:int=100, k=1, e=1):
    # A long form version of the above, just for clarity and testing
    assert k==1
    if not s or s.get('dy'):
        s = {'prev_y': y,
             'dy': {}}
        return y, np.eye(len(y)), s
    else:
        dy = y - s['prev_y']
        dy_hat, dy_cov, s['dy'] = buf_emp_pcov_d0_factory(y=dy, s=s['dy'], n_buffer=n_buffer)
        y['prev_y'] = y
        x = s['prev_y'] + dy_hat
        return x, dy_cov, s
    

BUF_EMP_D0_SKATERS = [buf_emp_pcov_d0_n20, buf_emp_pcov_d0_n50, buf_emp_pcov_d0_n100]
BUF_EMP_D1_SKATERS = [buf_emp_pcov_d1_n20, buf_emp_pcov_d1_n100]
BUFFERED_EMPIRICAL_SKATERS = BUF_EMP_D0_SKATERS + BUF_EMP_D1_SKATERS